from pydantic import BaseModel, ValidationError, validator
from collections import OrderedDict
import icsr.utils as utils
from typing import Union, Optional

# Structure for the model outputs:
class Evidence(BaseModel):
    span_start: int
    span_end: int
    span_str: str

    @validator("span_str")
    def span_logic(cls, field_value, values, field, config):
        span_start = values["span_start"]
        span_end = values["span_end"]
        span_str = field_value

        if span_end <= span_start:
            raise ValueError(
                f"span_end ({span_end}) should be greater than span_start ({span_start})."
            )

        if len(span_str) != span_end - span_start:
            raise ValueError(
                f"Expected span_str to be of length {span_end - span_start}, got length {len(span_str)}."
            )
        return span_str

    def __str__(self):
        return f"""'{self.span_str}'({self.span_start}-{self.span_end})"""

    def __eq__(self, other):
        return (
            (self.span_start == other.span_start)
            & (self.span_end == other.span_end)
            & (self.span_str == other.span_str)
        )

    def __hash__(self):
        return hash(str(self))


class Icsr(BaseModel):
    subject: str
    drug: str
    reaction: str
    subject_evidence: Optional[list[Evidence]]
    drug_evidence: Optional[list[Evidence]]
    reaction_evidence: Optional[list[Evidence]]

    @validator("subject", "drug", "reaction")
    def str_valid(cls, v):
        v = utils.normalize_text(v)
        if v == "":
            raise ValueError("ICSR attribute may not be empty after normalization.")
        return v

    @validator("subject_evidence", "drug_evidence", "reaction_evidence")
    def no_duplicate_evidence(cls, v):
        # deduplicate list but keep order
        v = list(OrderedDict.fromkeys(v))
        return v

    def has_valid_evidence(self):
        return (
            (len(self.subject_evidence) > 0)
            & (len(self.drug_evidence) > 0)
            & (len(self.reaction_evidence) > 0)
        )

    def __str__(self):
        return f"""ICSR:
    subject: {self.subject}
    subject_evidence: {self.subject_evidence}
    drug: {self.drug}
    drug_evidence: {self.drug_evidence}
    reaction: {self.reaction}
    reaction_evidence: {self.reaction_evidence}"""

    def __eq__(self, other):
        return (
            (self.subject == other.subject)
            & (self.drug == other.drug)
            & (self.reaction == other.reaction)
            & (self.subject_evidence == other.subject_evidence)
            & (self.drug_evidence == other.drug_evidence)
            & (self.reaction_evidence == other.reaction_evidence)
        )

    def __hash__(self):
        return hash(str(self))


class Document(BaseModel):
    icsrs: list[Icsr]
    identifier: str

    @validator("icsrs")
    def no_duplicate_icsrs(cls, v):
        # deduplicate list but keep order
        v = list(OrderedDict.fromkeys(v))
        return v

    def __getitem__(self, key):
        return self.icsrs[key]

    def __len__(self):
        return len(self.icsrs)

    def get_valid_icsrs(self):
        return Document(
            identifier=self.identifier,
            icsrs=[icsr for icsr in self.icsrs if icsr.has_valid_evidence()],
        )


class Dataset(BaseModel):
    # NOTE: rethink this
    documents: Union[dict[str, Document], list[Document]]
    # documents: dict[str, Document]

    @validator("documents")
    def dataset_not_empty(cls, v):
        return utils.dataset_not_empty(cls, v)

    def __getitem__(self, key):
        return self.documents[key]

    def __len__(self):
        return len(self.documents)

    def keys(self):
        return self.documents.keys()

    def values(self):
        return self.documents.values()

    def items(self):
        return self.documents.items()

    def get_valid_icsrs(self):
        return Dataset(
            documents={
                identifier: doc.get_valid_icsrs() for identifier, doc in self.items()
            }
        )


# Structure for the model inputs:
class TextDocument(BaseModel):
    text: str
    identifier: str

    @validator("text", "identifier")
    def str_not_empty(cls, v):
        if len(v) == 0:
            raise ValueError("Empty string not allowed.")
        return v


class TextDataset(BaseModel):
    documents: Union[dict[str, TextDocument], list[TextDocument]]

    @validator("documents")
    def dataset_not_empty(cls, v):
        return utils.dataset_not_empty(cls, v)

    def __getitem__(self, key):
        return self.documents[key]

    def __len__(self):
        return len(self.documents)

    def keys(self):
        return self.documents.keys()

    def values(self):
        return self.documents.values()

    def items(self):
        return self.documents.items()



import re
import string


# src: https://qa.fastforwardlabs.com/no%20answer/null%20threshold/bert/distilbert/exact%20match/f1/robust%20predictions/2020/06/09/Evaluating_BERT_on_SQuAD.html
def normalize_text(s):
    """Removing articles and punctuation, and standardizing whitespace are all typical text processing steps."""

    def remove_articles(text):
        regex = re.compile(r"\b(a|an|the)\b", re.UNICODE)
        return re.sub(regex, " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1(prec, rec):
    return (2 * (prec * rec) / (prec + rec)) if (prec + rec) > 0 else 0


def dataset_not_empty(cls, v):
    if len(v) == 0:
        raise ValueError("Dataset must not be empty.")
    if isinstance(v, list):
        dct_v = {}
        for doc in v:
            if doc.identifier in dct_v:
                raise ValueError(
                    f"Got duplicated document identifier: {doc.identifier}."
                )
            else:
                dct_v.update({doc.identifier: doc})
        v = dct_v
    return v



from pydantic import BaseModel, validator
from collections import OrderedDict
import icsr.structure as structure
import icsr.utils as utils
import re
from typing import Union


class EvidenceStrategy(BaseModel):
    return_all_evidence: bool = True

    def get_evidence(
        self,
        candidate: Union[structure.Icsr, structure.Document, structure.Dataset],
        support: Union[str, structure.TextDocument, structure.TextDataset],
    ) -> Union[structure.Icsr, structure.Document, structure.Dataset]:
        if isinstance(candidate, structure.Icsr):
            if not isinstance(support, str):
                raise TypeError(
                    f'Wrong support type. Expected "str", got {type(support)}.'
                )
            copy_candidate = candidate.copy(deep=True)
            self._icsr_evidence(copy_candidate, support)
            return copy_candidate

        if isinstance(candidate, structure.Document):
            if not isinstance(support, structure.TextDocument):
                raise TypeError(
                    f'Wrong support type. Expected "structure.TextDocument", got {type(support)}.'
                )
            copy_candidate = candidate.copy(deep=True)
            self._document_evidence(copy_candidate, support)
            return copy_candidate

        if isinstance(candidate, structure.Dataset):
            if not isinstance(support, structure.TextDataset):
                raise TypeError(
                    f'Wrong support type. Expected "structure.TextDataset", got {type(support)}.'
                )
            copy_candidate = candidate.copy(deep=True)
            self._dataset_evidence(copy_candidate, support)
            return copy_candidate

    def _icsr_evidence(self, icsr: structure.Icsr, text: str) -> None:
        icsr.subject_evidence = self.get_subject_evidence(icsr.subject, text)
        icsr.drug_evidence = self.get_drug_evidence(icsr.drug, text)
        icsr.reaction_evidence = self.get_reaction_evidence(icsr.reaction, text)

    def _document_evidence(
        self, document: structure.Document, text_doc: structure.TextDocument
    ) -> None:
        # check validity of document
        if text_doc.identifier != document.identifier:
            raise ValueError(f"Document and TextDocument identifiers should be equal.")
        # get evidence for every icsr
        for icsr in document.icsrs:
            text = text_doc.text
            self._icsr_evidence(icsr, text)

    def _dataset_evidence(
        self, dataset: structure.Dataset, text_dataset: structure.TextDataset
    ) -> None:
        # check dataset validity
        if text_dataset.keys() != dataset.keys():
            raise ValueError(f"Dataset and TextDataset identifiers should be equal.")
        # get evidence for every document
        for iden, doc in dataset.items():
            text_doc = text_dataset[iden]
            self._document_evidence(doc, text_doc)

    # functions that need to be implemented by an inheritant
    def get_subject_evidence(self, string: str) -> list[structure.Evidence]:
        raise NotImplementedError

    def get_drug_evidence(self, string: str) -> list[structure.Evidence]:
        raise NotImplementedError

    def get_reaction_evidence(self, string: str) -> list[structure.Evidence]:
        raise NotImplementedError


class ExactEvidence(EvidenceStrategy):
    def _get_all_evidence(self, string: str, text: str) -> list[structure.Evidence]:
        matches = list(re.finditer(string, text))
        if not self.return_all_evidence:
            matches = matches[:1]
        return [
            structure.Evidence(
                span_start=m.span()[0], span_end=m.span()[1], span_str=m.group()
            )
            for m in matches
        ]

    def get_subject_evidence(self, string: str, text: str) -> list[structure.Evidence]:
        return self._get_all_evidence(string, text)

    def get_drug_evidence(self, string: str, text: str) -> list[structure.Evidence]:
        return self._get_all_evidence(string, text)

    def get_reaction_evidence(self, string: str, text: str) -> list[structure.Evidence]:
        return self._get_all_evidence(string, text)



from pydantic import BaseModel, validator
from typing import Optional, Union
import numpy as np
from scipy.optimize import linear_sum_assignment

import icsr.structure as structure
import icsr.utils as utils

class SimilarityOutput(BaseModel):
    precision: float
    recall: float
    f1: float
    pred_support: Optional[int]
    gold_support: Optional[int]

    def to_array(self) -> np.array:
        return np.array(
            [
                self.precision,
                self.recall,
                self.f1,
            ]
        )

    @validator("precision", "recall", "f1")
    def format_metrics(v):
        return round(v, 4)


# TODO: implement a list of metrics and display them
class SimilarityOverview(BaseModel):
    pass


class Similarity(BaseModel):
    caefe_always_f1_alignment: Optional[bool] = True
    document_level_averaging: Optional[bool] = False
    count_zero_icsr: Optional[bool] = False
    string_similarity_method: str = "token"

    @validator("string_similarity_method")
    def check_string_similarity_method(cls, v):
        possible_values = ["token", "exact"]
        if v not in possible_values:
            raise ValueError(
                f"Expected 'string_similarity_method' to have a value in {possible_values}, but got value '{v}'."
            )
        return v

    def __call__(
        self,
        pred: Union[structure.Icsr, structure.Document, structure.Dataset],
        gold: Union[structure.Icsr, structure.Document, structure.Dataset],
    ) -> SimilarityOutput:
        if type(pred) is not type(gold):
            raise TypeError(
                f"Predicted and gold inputs should be the same type, got {type(pred)} and {type(gold)} respectively."
            )

        if not isinstance(
            pred, (structure.Icsr, structure.Document, structure.Dataset)
        ):
            raise TypeError(f"Invalid input type, got {type(pred)}.")

        if isinstance(pred, structure.Icsr):
            return self._icsr_similarity(pred, gold)

        if isinstance(pred, structure.Document):
            return self._document_similarity(pred, gold)

        if isinstance(pred, structure.Dataset):
            return self._dataset_similarity(pred, gold)

    def _string_similarity(self, pred_str: str, gold_str: str) -> SimilarityOutput:
        if self.string_similarity_method == "token":
            return self._string_token_similarity(pred_str, gold_str)
        elif self.string_similarity_method == "exact":
            return self._string_exact_similarity(pred_str, gold_str)

    def _string_token_similarity(
        self, pred_str: str, gold_str: str
    ) -> SimilarityOutput:
        # word tokenization
        # TODO: also split on '/'?
        pred_words = pred_str.split()
        gold_words = gold_str.split()

        # calculate word overlap metrics
        common_words = set(pred_words) & set(gold_words)

        # because of previous validation, we never have zero words
        prec = len(common_words) / len(pred_words)
        rec = len(common_words) / len(gold_words)
        f1 = utils.f1(prec, rec)

        return SimilarityOutput(precision=prec, recall=rec, f1=f1)

    def _string_exact_similarity(
        self, pred_str: str, gold_str: str
    ) -> SimilarityOutput:
        exact_match = pred_str == gold_str

        return SimilarityOutput(
            precision=exact_match, recall=exact_match, f1=exact_match
        )

    def _icsr_similarity(
        self, pred_icsr: structure.Icsr, gold_icsr: structure.Icsr
    ) -> SimilarityOutput:
        # subject
        subject_sim = self._string_similarity(pred_icsr.subject, gold_icsr.subject)

        # drug
        drug_sim = self._string_similarity(pred_icsr.drug, gold_icsr.drug)

        # reaction
        reaction_sim = self._string_similarity(pred_icsr.reaction, gold_icsr.reaction)

        # equal weighting of three similarities
        prec = (subject_sim.precision + drug_sim.precision + reaction_sim.precision) / 3
        rec = (subject_sim.recall + drug_sim.recall + reaction_sim.recall) / 3
        f1 = utils.f1(prec, rec)

        return SimilarityOutput(precision=prec, recall=rec, f1=f1)

    def _document_similarity(
        self, pred_document: structure.Document, gold_document: structure.Document
    ) -> SimilarityOutput:
        num_pred_icsrs = len(pred_document)
        num_gold_icsrs = len(gold_document)

        # both documents contain no icsrs
        if num_pred_icsrs == 0 and num_gold_icsrs == 0:
            if self.count_zero_icsr:
                return SimilarityOutput(
                    precision=1.0, recall=1.0, f1=1.0, pred_support=1, gold_support=1
                )
            else:
                return SimilarityOutput(
                    precision=1.0, recall=1.0, f1=1.0, pred_support=0, gold_support=0
                )
        # one document without icsrs
        elif num_pred_icsrs == 0 or num_gold_icsrs == 0:
            # for count_zero_icsr, we need to set the minimum num to 1
            if self.count_zero_icsr:
                return SimilarityOutput(
                    precision=0.0,
                    recall=0.0,
                    f1=0.0,
                    pred_support=max(1, num_pred_icsrs),
                    gold_support=max(1, num_gold_icsrs),
                )
            else:
                return SimilarityOutput(
                    precision=0.0,
                    recall=0.0,
                    f1=0.0,
                    pred_support=num_pred_icsrs,
                    gold_support=num_gold_icsrs,
                )

        # both documents contain icsrs
        # get unnormalized similarities for each alignment (prec / rec / f1 alignment)
        unnormalized_prec, unnormalized_rec = self._unnormalized_caefe_similarity(
            pred_document, gold_document
        )

        # normalize similarity scores
        prec = unnormalized_prec / num_pred_icsrs
        rec = unnormalized_rec / num_gold_icsrs
        f1 = utils.f1(prec, rec)

        return SimilarityOutput(
            precision=prec,
            recall=rec,
            f1=f1,
            pred_support=num_pred_icsrs,
            gold_support=num_gold_icsrs,
        )

    def _dataset_similarity(
        self, pred_dataset: structure.Dataset, gold_dataset: structure.Dataset
    ) -> SimilarityOutput:
        # check if the dataset contains the same documents
        if set(pred_dataset.keys()) != set(gold_dataset.keys()):
            raise ValueError(
                "Document identifiers in predicted and gold datasets should be equal."
            )

        # get similarity scores for each document
        document_similarities = {}
        for doc_id in pred_dataset.keys():
            document_similarities.update(
                {
                    doc_id: self._document_similarity(
                        pred_dataset[doc_id],
                        gold_dataset[doc_id],
                    )
                }
            )

        # average similarities per metric across the documents
        # TODO: this can be one code-block
        if self.document_level_averaging:
            prec_nom = sum([sim.precision for sim in document_similarities.values()])
            rec_nom = sum([sim.recall for sim in document_similarities.values()])
            prec_denom = len(document_similarities)
            rec_denom = len(document_similarities)
        # average similarities per metrics across ICSRs
        else:
            prec_nom = sum(
                [
                    sim.precision * sim.pred_support
                    for sim in document_similarities.values()
                ]
            )
            rec_nom = sum(
                [
                    sim.recall * sim.gold_support
                    for sim in document_similarities.values()
                ]
            )
            prec_denom = sum(
                [sim.pred_support for sim in document_similarities.values()]
            )
            rec_denom = sum(
                [sim.gold_support for sim in document_similarities.values()]
            )

        prec = prec_nom / prec_denom
        rec = rec_nom / rec_denom
        f1 = utils.f1(prec, rec)

        return SimilarityOutput(
            precision=prec,
            recall=rec,
            f1=f1,
            pred_support=prec_denom,
            gold_support=rec_denom,
        )

    def _unnormalized_caefe_similarity(
        self, pred_document: structure.Document, gold_document: structure.Document
    ) -> tuple[float, float]:
        """Computes the un-normalized Constrained Entity-Alignment F-Measure (CEAF) for evaluating coreference.

        Gold and predicted mentions are aligned into entityings which maximise a similarity metric
        https://www.semanticscholar.org/paper/On-Coreference-Resolution-Performance-Metrics-Luo/de133c1f22d0dfe12539e25dda70f28672459b99

        """
        # calcalate pairwise icsr similarities
        scores = np.zeros((len(pred_document), len(gold_document), 3))
        for i in range(len(pred_document)):
            for j in range(len(gold_document)):
                pred_icsr = pred_document[i]
                gold_icsr = gold_document[j]
                scores[i, j] = self._icsr_similarity(pred_icsr, gold_icsr).to_array()

        # if we are using the f1 alignment
        if self.caefe_always_f1_alignment:
            prec_alignment = linear_sum_assignment(scores[:, :, 2], maximize=True)
            rec_alignment = prec_alignment
        # use different prec and rec alignments
        else:
            prec_alignment = linear_sum_assignment(scores[:, :, 0], maximize=True)
            rec_alignment = linear_sum_assignment(scores[:, :, 1], maximize=True)

        # unnormalized prec and rec
        unnormalized_prec = sum(scores[prec_alignment[0], prec_alignment[1], 0])
        unnormalized_rec = sum(scores[rec_alignment[0], rec_alignment[1], 1])

        return unnormalized_prec, unnormalized_rec
